#!/usr/bin/perl -w

use strict;
use warnings;
use Data::Dumper;


my $usage = << "USE";
	
	Join results for large chromosomes.
	
	Usage: fasta joiner
	perl ngm_joiner.pl <sam file> <equal scoring limit>
USE

my $argcount = $#ARGV + 1;
my $reqArgs = 2;
$argcount == $reqArgs or die "$usage";

my $file_path = $ARGV[0];
my $max_equal = $ARGV[1];

open(FHI, "<$file_path") or die "Unable to open file $file_path \n";

my $hit_list;
my $hit_index = 0;
my $last_id = 0;

while(<FHI>) {
	if(/^[^@]/) {
		chomp($_);
		my @tmp = split(/\t/, $_);
		if(length($last_id) > 0 && $hit_index > 0 && $tmp[0] ne $hit_list->[$hit_index - 1][0]) {
			#New read found
			findBestHit();
			
			#Clear hit  list
			@{$hit_list} = ();
			$hit_index = 0;
		}
		#Add to hit list
		$hit_list->[$hit_index++] = \@tmp;
	}	

}
findBestHit();
close(FHI);

sub findBestHit {
	#print Dumper($hit_list);
	if($hit_index > 1) {
		#Find max score
		my $max_score = 0;
		for(my $i = 0; $i < $hit_index; ++$i) {
			my $score = int(substr($hit_list->[$i][11], 5));
			if($score > $max_score) {
				$max_score = $score;
			}
		}
		#Print best scoring hits
		my $equal = 0;
		my %pos_list = ();
		for(my $i = 0; $i < $hit_index && $equal < $max_equal; ++$i) {
			my $score = int(substr($hit_list->[$i][11], 5));
			if($score == $max_score) {
				my $pos = $hit_list->[$i][3];
				if(!$pos_list{$pos}) {
					$pos_list{$pos} = 1;
					printHit($i);
					$equal += 1;
				} else {
					#print STDERR "Ignoring position $pos.\n";
				}
			}
		}			
	} else {
		printHit(0);
	}			
}

sub printHit {
	my $i = $_[0];
	my $chr = $hit_list->[$i][2];
	#Correct position of hits from split chromosomes
	if($chr =~ /_([0-9]+)$/) {
		my $position = int($1);
		$chr =~ /(.*?)_[0-9]+$/;
		$hit_list->[$i][2] = $1;
		$hit_list->[$i][3] = int($hit_list->[$i][3]) + $position;				
	} 
	print join("\t", @{$hit_list->[$i]}) . "\n";
}

