/* Simple API for FASTA file reading
 * for Bio5495/BME537 Computational Molecular Biology
 * SRE, Sun Sep  8 05:35:11 2002 [AA2721, transatlantic]
 * CVS $Id$
 */
// qx: src from http://www.cse.msu.edu/~cse891/Sect001/hw1-codes

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <iostream>

#include "FastqParser.h"

//int FastqParser::ReadFASTA(FASTAFILE *ffp, string & name, char **ret_seq,
//		FilePosition *retNamePos, int *retNameLen, int *ret_L) {
//	char *s;
//	int n;
//
//	/* Peek at the lookahead buffer; see if it appears to be a valid FASTA descline.
//	 */
//	if (ffp->buffer[0] != '@')
//		return 0;
//
//	name = (ffp->buffer + 1);
//	/* Parse out the name: the first non-whitespace token after the >
//	 */
//	//s = strtok(ffp->buffer + 1, " \t\n");
//	//name = (char*) malloc(sizeof(char) * (strlen(s) + 1));
//	//strcpy(name, s);
//	//Ignore > and \n
//	*retNamePos = ffp->filePos + 1;
//	*retNameLen = strlen(ffp->buffer) - 2;
//
//	/* Everything else 'til the next descline is the sequence.
//	 * Note the idiom for dynamic reallocation of seq as we
//	 * read more characters, so we don't have to assume a maximum
//	 * sequence length.
//	 */
//	n = 0;
//	bool isSeqLine = true;
//	ffp->filePos = ftell(ffp->fp);
//	while (fgets(ffp->buffer, FASTA_MAXLINE, ffp->fp)) {
//		if (ffp->buffer[0] == '@')
//			break; /* a-ha, we've reached the next descline */
//		if((ffp->buffer[0] == '+')) {
//		isSeqLine = isSeqLine && !(ffp->buffer[0] == '+');
//		if (isSeqLine) {
//			for (s = ffp->buffer; *s != '\0'; s++) {
//				if (!isalpha(*s))
//					continue; /* accept any alphabetic character */
//
//				seq[n] = *s; /* store the character, bump length n */
//				n++;
//				if (nalloc == n) /* are we out of room in seq? if so, expand */
//				{ /* (remember, need space for the final '\0')*/
//					nalloc += 128;
//					seq = (char*) realloc(seq, sizeof(char) * nalloc);
//				}
//			}
//		} else {
//
//		}
//		ffp->filePos = ftell(ffp->fp);
//	}
//	seq[n++] = '\0';
//
//	*ret_seq = (char*) malloc(sizeof(char) * n);
//	//*ret_name = name;
//	strcpy(*ret_seq, seq);
//	*ret_L = n - 1;
//	return 1;
//}

//int FastqParser::getMaxReadLength(int & readCount, FilePosition & fileSize) {
//	FILE * fp = ffp->fp;
//
//	if (fp == NULL) {
//		return -1;
//	}
//	readCount = 0;
//	FilePosition currentPos = ftell(fp);
//	fseek(fp, 0, SEEK_SET);
//	char buffer[FASTA_MAXLINE];
//	int maxReadLength = 0;
//	int currentLength = 0;
//	bool isSeqLine = true;
//	while ((fgets(buffer, FASTA_MAXLINE, fp))) {
//		if (buffer[0] == '@') {
//			readCount += 1;
//			maxReadLength = std::max(currentLength, maxReadLength);
//			currentLength = 0;
//			isSeqLine = true;
//		} else {
//			isSeqLine = isSeqLine && buffer[0] != '+';
//			if (isSeqLine) {
//				int length = strlen(buffer);
//				currentLength += ((buffer[length - 1] == '\n') ? length - 1 : length);
//			}
//		}
//	}
//	maxReadLength = std::max(currentLength, maxReadLength);
//	fileSize = ftell(fp);
//	fseek(fp, currentPos, SEEK_SET);
//	return maxReadLength;
//}

int FastqParser::getMaxReadLength(int & readCount, FilePosition & fileSize) {
	FILE * fp = ffp->fp;

	if (fp == NULL) {
		return -1;
	}
	readCount = 0;
	FilePosition currentPos = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	char buffer[FASTA_MAXLINE];
	int maxReadLength = 0;

	int state = 0;

	while ((fgets(buffer, FASTA_MAXLINE, fp))) {
		if (strlen(buffer) > 1) {
			switch (state) {
			case 0:
				if (buffer[0] != '@') {
					throw "File corrupted";
				}
				state += 1;
				readCount += 1;
				break;
			case 1:
				maxReadLength = std::max(maxReadLength, (int) strlen(buffer));
				state += 1;
				break;
			case 2:
				if ((buffer[0] != '+')) {
					//TODO: throw exception
					throw "File corrupted";
				} else {
					state += 1;
				}
				break;
			case 3:
				state = 0;
				break;
			}
		}
	}
	fileSize = ftell(fp);
	fseek(fp, currentPos, SEEK_SET);
	return maxReadLength;
}

char const * FastqParser::getFormatName() {
	return "fastq";
}

//TODO: remove
#include <iostream>
using std::cout;

bool FastqParser::read(string & name, Entry * entry) {

	char *s;
	int n;

	//	std::cout << "Name: " << ffp->buffer << std::endl;
	if (ffp->buffer[0] != '@')
		return false;

	//	char * debugName = new char[500];
	name = (ffp->buffer + 1);
	entry->nameOffset = ffp->filePos + 1;
	entry->nameLength = strlen(ffp->buffer) - 2;
	//	strcpy(debugName, ffp->buffer);

	entry->qualityLength = 0;

	n = 0;
	bool foundNext = false;
	ffp->filePos = ftell(ffp->fp);

	int state = 1;
	while (!foundNext && fgets(ffp->buffer, FASTA_MAXLINE, ffp->fp)) {
		//		std::cout << "State: " << state << std::endl;
		if (strlen(ffp->buffer) > 1) {
			switch (state) {
			case 1:
				for (s = ffp->buffer; *s != '\0' && *s != '\n'; s++) {
					//				if (!isalpha(*s))
					//					continue; /* accept any alphabetic character */

					seq[n] = *s; /* store the character, bump length n */
					n++;
					if (nalloc == n) /* are we out of room in seq? if so, expand */
					{ /* (remember, need space for the final '\0')*/
						nalloc += 128;
						//cout << nalloc << " " << sizeof(char) * nalloc << "\n";
						seq = (char*) realloc(seq, sizeof(char) * nalloc);
					}
				}
				seq[n++] = '\0';
//				cout << "xx: " << n << " " << strlen(seq) << "\n";
				entry->sequence = (char*) malloc(sizeof(char) * (n));
				strcpy(entry->sequence, seq);
				entry->sequenceLength = n - 1;
				state += 1;
				break;
			case 2:
				if ((ffp->buffer[0] != '+')) {
					//TODO: throw exception
					throw "File corrupted";
				} else {
					state += 1;
				}
				break;
			case 3:
				entry->qualityOffset = ffp->filePos;
				entry->qualityLength = strlen(ffp->buffer) - 1;
				state += 1;
				break;
			case 4:
				foundNext = (ffp->buffer[0] == '@');
				if (!foundNext) {
					throw "File corrupted";
				} else {
					return true;
				}
				break;
			}
		}
		ffp->filePos = ftell(ffp->fp);
	}
	//	if (state != 4) {
	//		throw debugName;
	//	} else {
	//		delete[] debugName;
	//		debugName = 0;
	//	}

	//entry->qualityLength = n - 1;
	return true;
}

//bool FastqParser::read(string & name, Entry * entry) {
//
//	char *s;
//	int n;
//
//	if (ffp->buffer[0] != '@')
//		return false;
//
//	char * debugName = new char[500];
//	name = (ffp->buffer + 1);
//	entry->nameOffset = ffp->filePos + 1;
//	entry->nameLength = strlen(ffp->buffer) - 2;
//	strcpy(debugName, ffp->buffer);
//
//	entry->qualityLength = 0;
//
//	n = 0;
//	bool isSeqLine = true;
//	bool isLastLine = true;
//	ffp->filePos = ftell(ffp->fp);
//	while (fgets(ffp->buffer, FASTA_MAXLINE, ffp->fp)) {
//		if (ffp->buffer[0] == '@') {
//			isLastLine = false;
//			break;
//		}
//		if ((ffp->buffer[0] == '+')) {
//			isSeqLine = false;
//			entry->qualityOffset = ffp->filePos + 2;
//		}
//		//isSeqLine = isSeqLine && !(ffp->buffer[0] == '+');
//		if (isSeqLine) {
//			for (s = ffp->buffer; *s != '\0' && *s != '\n'; s++) {
//				//				if (!isalpha(*s))
//				//					continue; /* accept any alphabetic character */
//
//				seq[n] = *s; /* store the character, bump length n */
//				n++;
//				if (nalloc == n) /* are we out of room in seq? if so, expand */
//				{ /* (remember, need space for the final '\0')*/
//					nalloc += 128;
//					seq = (char*) realloc(seq, sizeof(char) * nalloc);
//				}
//			}
//		}
//		ffp->filePos = ftell(ffp->fp);
//	}
//	entry->qualityLength = ffp->filePos - entry->qualityOffset
//			- ((isLastLine) ? 0 : 1);
//	seq[n++] = '\0';
//
//	entry->sequence = (char*) malloc(sizeof(char) * n);
//	strcpy(entry->sequence, seq);
//	entry->sequenceLength = n - 1;
//	if(strlen(seq) <= 0 || entry->sequenceLength <= 0) {
//		std::cout << debugName << std::endl;
//	}
//	//entry->qualityLength = n - 1;
//	return true;
//}
