/*
 * main.cpp
 *
 *  Created on: Feb 16, 2011
 *      Author: philipp_
 */
#include <iostream>
#include <string>
#include <fstream>
#include <stdlib.h>

#include <errno.h>
#include <string.h>

#include "Ngmc.h"
#include "NgmRefWriter.h"
#include "NgmReadWriter.h"

#include "Statistics.h"
#include "Length_Statistics.h"
#include "Composition_Statistics.h"

#ifndef NDEBUG
bool cDebug = true;
#else
bool cDebug = false;
#endif

using std::cout;
using std::endl;
using std::ofstream;

char const * version = "1.1.1";

vector<Statistics *> init_jops(int mode, int maxReadLength, char const * const fileName) {
	vector<Statistics *> jops;
	switch (mode) {
		case 1:
			jops.push_back(new Composition_Statitstics(maxReadLength, fileName));
			break;
		case 2:
			jops.push_back(new Length_Statistics(maxReadLength, fileName));
			break;
		case 3:
			jops.push_back(new Composition_Statitstics(maxReadLength, fileName));
			jops.push_back(new Length_Statistics(maxReadLength, fileName));
			break;
	}

	return jops;
}

int main(int argc, char **argv) {

	char const * arch[] = { "x86", "x64" };
	char const * build = (cDebug) ? " (DEBUG)" : "";
	//cout << "Startup : " << arch[sizeof(void*) / 4 - 1] << build << " (build " << __DATE__ << " " << __TIME__ << ")" << endl;

	int readCount = -1;
	if (argc > 3) {
		try {
			bool paired = argc > 4;
			int mode = atoi(argv[1]);
			int pre_mode = atoi(argv[2]);
			Parser * parser = getParser(argv[3]);
			Parser * parserRight = (paired) ? getParser(argv[4]) : 0;

			if (parser != 0 && (!paired || parserRight != 0)) {

				parser->open(argv[3]);
				//cout << "Version: " << version << endl;
				//cout << "Processing " << argv[3] << endl;
				cout << "Format: " << parser->getFormatName() << endl;

				if (mode == 0) {// jUST READS
					if (!paired) {
						FilePosition fileSize;
						int maxReadLength = parser->getMaxReadLength(readCount, fileSize) | 1;
						cout << "Reads found: " << readCount << endl;
						cout << "Max. read length: " << (maxReadLength + 1) << endl;
						cout << "Converting reads..." << endl;

						vector<Statistics *> jops = init_jops(pre_mode, maxReadLength, argv[3]);

						NgmReadWriter writer(argv[3]);
						int readsWritten = convert(parser, writer, maxReadLength, readCount, jops);

						for (vector<Statistics *>::iterator i = jops.begin(); i != jops.end(); i++) {
							(*(*i)).print();
							delete (*i);
						}

						jops.clear();
						cout << "Reads written: " << readsWritten << endl;
						cout << "Reads skipped: " << (readCount - readsWritten) << endl;
					} else {
						parserRight->open(argv[4]);
						cout << "Processing " << argv[4] << endl;
						cout << "Format: " << parserRight->getFormatName() << endl;

						int readCountLeft = 0;
						int readCountRight = 0;
						FilePosition fileSizeLeft;
						FilePosition fileSizeRight;
						int maxReadLengthLeft = parser->getMaxReadLength(readCountLeft, fileSizeLeft) | 1;
						int maxReadLengthRight = parserRight->getMaxReadLength(readCountRight, fileSizeRight) | 1;
						int maxReadLength = std::max(maxReadLengthLeft, maxReadLengthRight);
						readCount = (readCountLeft + readCountRight);

						cout << "Reads found: " << readCount << endl;
						cout << "Max. read length: " << (maxReadLength + 1) << endl;
						cout << "Converting reads..." << endl;
						NgmReadWriter writerLeft(argv[3]);
						NgmReadWriter writerRight(argv[4]);

						vector<Statistics*> jops_left = init_jops(pre_mode, maxReadLength, argv[3]);
						vector<Statistics*> jops_right = init_jops(pre_mode, maxReadLength, argv[4]);

						int readsWritten = convertPairs(parser, parserRight, writerLeft, writerRight, maxReadLength, maxReadLength, (fileSizeLeft + fileSizeRight) % readCountLeft + 1, jops_left, jops_right);

						for (size_t i = 0; i < jops_left.size(); i++) {
							(*jops_left[i]).print();
							delete jops_left[i];
							(*jops_right[i]).print();
							delete jops_right[i];
						}

						jops_left.clear();
						jops_right.clear();
						cout << "Reads written: " << readsWritten << endl;
						cout << "Reads skipped: " << (readCount - readsWritten) << endl;
					}
				} else {
					FilePosition fileSize;
					int maxReadLength = parser->getMaxReadLength(readCount, fileSize);
					cout << "Reference sequences found: " << readCount << endl;
					cout << "Max. sequence length: " << (maxReadLength + 1) << endl;
					cout << "Converting sequences..." << endl;
					NgmRefWriter writer(argv[3]);
					int readsWritten = convertRef(parser, writer, mode, readCount);

					cout << "Sequences written: " << readsWritten << endl;
					cout << "Sequences skipped: " << (readCount - readsWritten) << endl;
				}

			}
			parser->close();
			delete parser;
			parser = 0;
		} catch (char const * ex) {
			std::cerr << "EXCEPTION: " << ex << std::endl;
			return 1;
		} catch (string ex) {
			std::cerr << "EXCEPTION: " << ex << std::endl;
			return 1;
		} catch (std::exception const &ex) {
			cout << "Error during file operation: " << endl << strerror(errno) << endl;
			return 1;
		}
	} else {
		cout << "NextGenMap Converter v " << version << endl;
		cout << "Missing arguments." << endl << endl;
		cout << "./NGMC n path  " << endl << endl;
		cout << "Parameter:  " << endl;
		cout << "n:\t\tLength of the longest read for reference sequences files" << endl;
		cout << "\t\t0 for read files." << endl;
		cout << "path:\t\tPath to a single file containing read files or reference sequences files" << endl << endl;

		cout << "Example:" << endl;
		cout << "./NGMC 0 reads.fa" << endl;
		cout << "./NGMC 83 reference.fa" << endl << endl;
		return 1;
	}
	return 0;
}
