#include "CS.h"
#include "NGM.h"

#include "CSCache.h"

#include "Timing.h"
#include <memory.h>
#include <stdlib.h>

#include "Debug.h"

#undef module_name
#define module_name "CS"

volatile int CS::s_ThreadCount = 0;

// Default-Werte
int RefEntry::MaxRefsPerEntry = 100;	// Anzahl Positionen, die pro RefEntry-Knoten gespeichert werden
int CS::cThreshhold = 4;
int CS::cSrchTableLen = 9999991;
int CS::cRefTableLen = 9999991;
uint CS::prefixBasecount = 10;
uint CS::prefixBits = prefixBasecount*2;
//uint CS::prefixTableEntries = 1 << prefixBits;
ulong CS::prefixMask = (1 << prefixBits) - 1;
uint CS::prefixBaseSkip = 0;

int const stageTableLen[8] = { 32749, 65521, 131071, 262139, 524287, 1048573, 2097143, 4194301 };

bool CS::useCache = false;
std::string CS::cacheFileName = "";

PrefixRefRelay * * RefTable = 0;

void InsertRef(int prefix, int pos);
void ExtractPrefix(char const * const  seq);


inline int min(int a, int b)
{
	return (a < b) ? a : b;
}

// A->0 C->1 T->2 G->3
inline char encode(char c)
{
	return (c >> 1) & 3;
}

//uint maxLen = 0;
//uint nextProgress = 0;

// Iteriert ueber jeden Praefix in sequence und fuehrt fuer diesen die Funktion func aus
void CS::PrefixIteration(char const * sequence, uint length, PrefixIterationFn func, void* data, uint offset)
{
//	if(maxLen == 0) {
//		maxLen = length;
//		nextProgress = (maxLen / 100);
//	}
	if (length < prefixBasecount)
		return;

	if (*sequence == 'N')
	{
		uint n_skip = 1;
		while (*(sequence + n_skip) == 'N')
			++n_skip;

		sequence += n_skip;
		length -= n_skip;
		offset += n_skip;
	}

	int prefix = 0;
	for (uint i = 0; i < prefixBasecount-1; ++i)
	{
		char c = *(sequence+i);
		if (c == 'N')
		{
			PrefixIteration(sequence+i+1, length-i-1, func, data, offset+i+1);
			return;
		}

		prefix = prefix << 2;
		char cx = encode(c);
		prefix |= cx;
	}

	uint skipcount = prefixBaseSkip;
	for (uint i = prefixBasecount-1; i < length; ++i)
	{
//		int step = (maxLen - length) + i;
//		if(step > nextProgress) {
//			int progress = step * 100.0f / maxLen;
//			Log.Message("Progress: %d %\r", progress);
//			nextProgress += (maxLen / 100);
//		}
		char c = *(sequence+i);
		if (c == 'N')
		{
			PrefixIteration(sequence+i+1, length-i-1, func, data, offset+i+1);
			return;
		}

		prefix = prefix << 2;
		char cx = encode(*(sequence+i));
		prefix |= cx;
		prefix &= prefixMask;

		if (skipcount == prefixBaseSkip)
		{
			func(prefix, offset+i+1-prefixBasecount, data);
			skipcount = 0;
		}
		else
		{
			++skipcount;
		}
	}
}

int lastSeqTotal = 0;

void CS::BuildPrefixTable(ulong prefix, int pos, void* data)
{
	/* // Since PrefixIteration masks, this can never happen
	if (!(prefix < prefixTableEntries))
	{
		Log.Error("Illegal prefix encountered while building prefix table.");
		Fatal();
	}*/
	NGM.Stats->CurrentBase = pos;
	NGM.Stats->CurrentTotal = lastSeqTotal + pos;

	int refSeqId = *(int*)data;
	
	SequenceLocation tmp = {pos, refSeqId};

	SaveToRefTable(prefix, tmp);
}

int locTotal = 0; // TODO remove
int usedTotal = 0;

void CS::PrefixSearch(ulong prefix, int pos, void* data)
{
	CS * cs = (CS*)data;
	int mode = 0;

	RefEntry * cur = cs->m_RefProvider->GetRefEntry(prefix);	// Liefert eine liste aller Vorkommen dieses Praefixes in der Referenz
	if (cur != 0 && cur->refTotal > 1000) // TODO: cs_discardPrefixWithRef
	{
		if (useCache)
			delete cur;
		cur = 0;
	}

	int n = 0;
	while (cur != 0)
	{
		//Log.Green("Qry Seq %i - Prefix 0x%x got %i locs (sum %i)", m_CurrentQrySeq, prefix, cur->refTotal, locTotal);
		locTotal += cur->refTotal; // TODO remove

		n += cur->refTotal;
		for (int i = 0; i < cur->refCount; ++i)
		{
			SequenceLocation curLoc = cur->ref[i];
			int poffs = cs->GetBin(curLoc.m_Location-pos);	// position offset

			if (mode != 0)
			{
				SequenceLocation tmp = {poffs, curLoc.m_RefId};
				if ((++cs->iTable[tmp]) == cThreshhold)
				{
					++cs->m_Candidates;
					//Log.Message("Threshhold hit at location %i", poffs);
				}
			}
			else
			{
				int hpo = hash(poffs);
				//       intelligentere overflow-detektion: irgendwas wie aktueller pr�fix �ber 50% des reads
				//       und f�llstand der tabelle mit pr�fixposition vergleichen -> wenn �ber einem threshhold,
				//       sofort abbrechen und auf fallback mode gehen
				int hpoc = cSrchTableLen;
				while ( (cs->rTable[hpo].Location.m_Location != cInvalidLocation) && !(cs->rTable[hpo].Location.m_Location == poffs && cs->rTable[hpo].Location.m_RefId == curLoc.m_RefId))
				{
					++hpo;
					if (hpo >= cSrchTableLen)
						hpo = 0;
					if (--hpoc == 0)
					{
						Log.Message("CS table overflow on <%i, %i> (prefix 0x%x at pos %i) seq #%i", poffs, curLoc.m_RefId, prefix, pos, cs->m_CurrentSeq);
						throw 1;
					}
				}
				
				if (cs->rTable[hpo].Location.m_Location == cInvalidLocation)
				{
					++usedTotal; // TODO remove
					cs->rTable[hpo].Location.m_Location = poffs;
					cs->rTable[hpo].Location.m_RefId = curLoc.m_RefId;
				}

				if (++cs->rTable[hpo].Score.i == cThreshhold)
				{
					++cs->m_Candidates;
				}
			}
		}

		RefEntry * last = cur;
		cur = cur->nextEntry;

		if (useCache)
		{
			delete last;
		}
	}
}

int CS::CollectResults(MappedRead * read, int mode, int qrynr)
{
#ifndef NDEBUG
	Log.Verbose("Qry #%i got %i results", qrynr, m_Candidates);
#endif
	if (m_Candidates == 0)
	{
		Log.Verbose("Unmapped read %i", qrynr);
		NGM.AddUnmappedRead(qrynr, MFAIL_NOCAND);
		if (mode == 0)
			for (int i=0; i < cSrchTableLen; ++i)
			{
				rTable[i] = sInvalidLocationScore;
			}
		return 1;
	}

	int n = 0;
	if (mode == 0)
	{
		for (int i = 0; i < cSrchTableLen; ++i)
		{
			if (rTable[i].Location.m_Location != cInvalidLocation)
			{
				if (rTable[i].Score.i >= cThreshhold)
				{
					rTable[i].Location.m_Location = ResolveBin(rTable[i].Location.m_Location);
					read->AddScore(rTable[i]);
					++n;
				}
			}
			rTable[i] = sInvalidLocationScore;
		}
	}
	else if (mode == 1)
	{
		for (std::map<SequenceLocation, int>::iterator itr = iTable.begin(); itr != iTable.end(); itr++)
		{
			//if (itr->first.m_Location >= 0)
			{
				if (itr->second >= cThreshhold)
				{
					LocationScore score = { itr->first, {itr->second}, read };

					score.Location.m_Location = ResolveBin(score.Location.m_Location);
					read->AddScore( score );
					//Log.Message("Adding location %i (score %i)", itr->first.m_Location, itr->second);
					++n;
				}
			}
		}
		iTable.clear();
	}
	if (n != m_Candidates)
	{
		Log.Error("Candidate list size mismatch on read %i (n = %i, m_C = %i)", m_CurrentSeq, n, m_Candidates);
	}

	return 0;
}

void CS::SendToBuffer(MappedRead * read)
{
	if (read == 0)
		return;

	int count = read->Scores.size();

	if (count < NGM.bCSSW.Capacity())
	{
		NGM.bCSSW.Write(&read->Scores[0], count);
		m_WrittenReads += count;
	}
	else
	{
		Log.Warning("Read %i discarded due to buffer overflow", read->ReadId);
		NGM.AddUnmappedRead(read->ReadId, MFAIL_BOFLOW);
		++m_DiscardedReads;
		delete read;
	}
}

void CS::Cleanup()
{
	/*if (RefTable != 0)
	{
		Log.Message("Clearing RefTable");
		for (uint i = 0; i < prefixTableEntries; ++i)
		{
			if (RefTable[i] != 0)
				delete RefTable[i];
			RefTable[i] = 0;
		}
		delete[] RefTable;
	}
	Log.Message("RefTable cleared");*/
}

void CS::RunBatch()
{
	m_ReadBatch = new MappedRead*[m_CurrentCount];

	// Initialize empty MappedRead objects for current batch
	for (int i = 0; i < m_CurrentCount; ++i)
	{
		m_ReadBatch[i] = new MappedRead(m_CurrentStart + i);
	}

	for (m_CurrentPage = 0; m_CurrentPage < m_RefProvider->GetPageCount(); ++m_CurrentPage)
	{
		Log.Message("CS Thread %i loading Page %i", m_CSThreadID, m_CurrentPage);
		m_RefProvider->LoadPage(m_CurrentPage);
		Log.Message("Load done");

		for (int i = 0; i < m_CurrentCount; ++i)
		{
			m_CurrentSeq = m_CurrentStart + i;
			if (m_CSThreadID < NGMStats::cMaxThreadStats)
				NGM.Stats->CS[m_CSThreadID].CurrentRead = m_CurrentSeq;

			char const * const qrySeq = SequenceProvider.GetQrySequence(m_CurrentSeq);
			int qryLen = strlen(qrySeq);
			try
			{
				if (cSrchTableLen == 0) // cTableLen 0 means always use std::map rather then hashing
					throw 2;

				m_Candidates = 0;
				//CheckRTable();	// Debug/Validation

				PrefixIteration(qrySeq, qryLen, &CS::PrefixSearch, this);

				if (m_Candidates == 0)
				{
					NGM.AddUnmappedRead(m_CurrentSeq, MFAIL_NOCAND);
					delete m_ReadBatch[i];
					m_ReadBatch[i] = 0;
				}
				else
				{
					if (NGM.Paired() && (i & 0x1))
					{
						m_ReadBatch[i]->Paired = m_ReadBatch[i-1];
						if (m_ReadBatch[i-1] != 0)
							m_ReadBatch[i-1]->Paired = m_ReadBatch[i];
					}

					CollectResults(m_ReadBatch[i], 0, m_CurrentSeq);
				}
			}
			catch (int overflow)
			{
				if (overflow == 1)
				{
					// Clear rTable
					for (int i = 0; i < cSrchTableLen; ++i)
						rTable[i] = sInvalidLocationScore;

					// TODO remove
					Log.Message("Total locations for read until interrupt: %i (%i candidates) %i used", locTotal, m_Candidates, usedTotal);
					locTotal = 0;
					usedTotal = 0;
				}
				m_Candidates = 0;

				// Discard read
				++m_DiscardedReads;
				NGM.AddUnmappedRead(m_CurrentSeq, MFAIL_CSOFLOW);

				delete m_ReadBatch[i];
				m_ReadBatch[i] = 0;

				// Initiate fallback search on iTable
				//PrefixIteration(qrySeq, qryLen, &CS::PrefixSearch, 1);
				//CollectResults(1, m_CurrentQrySeq);
			}
		}
	}

	for (int i = 0; i < m_CurrentCount; ++i)
	{
		SendToBuffer(m_ReadBatch[i]);
	}

	delete[] m_ReadBatch;
}

void CS::DoRun()
{
	Log.Message("Launching CS Thread %i (NGM Thread %i)", m_CSThreadID, m_TID);
	rTable = new LocationScore[cSrchTableLen];
	for (int i = 0; i < cSrchTableLen; ++i)
	{
		rTable[i] = sInvalidLocationScore;
	}

	int batchsize = 1000;		// default batchsize
	if (Config.Exists("cs_batchsize"))
		batchsize = Config.GetInt("cs_batchsize");

	if (Config.Exists("cs_threshhold"))
		cThreshhold = Config.GetInt("cs_threshhold");

	if (Config.Exists("cs_prefixskip"))
		prefixBaseSkip = Config.GetInt("cs_prefixskip", 0, -1);

	if (useCache)
		m_RefProvider = CSCache::LoadCache(cacheFileName);
	else
		m_RefProvider = this;

	Timer tmr;

	while ((m_CurrentCount = NGM.GetNextReadBatch(batchsize, m_CurrentStart)) != 0)
	{
		Log.Message("CS Thread %i got batch %i[+%i]", m_TID, m_CurrentStart, m_CurrentCount);
		tmr.ST();

		RunBatch();

		Log.Verbose("CS Thread %i finished batch %i[+%i] (elapsed: %.3fs)", m_TID, m_CurrentStart, m_CurrentCount, tmr.ET());
	}

	//Notice output buffer we're done
	NGM.bCSSW.Release();

	// Close cache if we're using one
	if (m_RefProvider != this)
		delete m_RefProvider;

	Log.Message("CS Thread %i finished (%i reads written, %i reads discarded)", m_TID, m_WrittenReads, m_DiscardedReads);
}

void CS::Init()
{
	if (Config.Exists("cs_maxRefsPerEntry"))
		RefEntry::MaxRefsPerEntry = Config.GetInt("cs_maxRefsPerEntry");

	if (Config.Exists("cs_reftablelen"))
		cRefTableLen = Config.GetInt("cs_reftablelen");

	if (Config.Exists("cs_searchTableLen"))
		cSrchTableLen = Config.GetInt("cs_searchTableLen");

	if (Config.Exists("cs_prefixsize"))
		prefixBasecount = Config.GetInt("cs_prefixsize", 4, 32);

	prefixBits = prefixBasecount*2;
	prefixMask = (1 << prefixBits) - 1;

	cacheFileName = std::string(Config.GetString("ref")) + ".cache";
	cacheFileName = cacheFileName.substr(cacheFileName.find_last_of("/\\")+1);

	if (Config.Exists("cs_cache"))
		useCache = (Config.GetInt("cs_cache") == 1);

	int cachePageLimit = 20000000;
	if (useCache && Config.Exists("cs_cachepagelimit"))
	{
		cachePageLimit = Config.GetInt("cs_cachepagelimit");
	}

	if (useCache && FileExists(cacheFileName.c_str()))
	{
		Log.Message("Prefix Table cached");
	}
	else
	{
		Timer init_tmr;
		init_tmr.ST();
		Log.Message("Building Prefix Table");


		CSCache * cache = 0;
		if (useCache)
			cache = CSCache::CreateCache(cacheFileName);

		NGM.Stats->CSStage = 1;
		CS CSInit = CS(false);

		RefTable = new PrefixRefRelay*[cRefTableLen];
		memset(RefTable, 0, sizeof(PrefixRefRelay*)*cRefTableLen);

		NGM.Stats->CSStage = 2;
		int ref_mode = Config.GetInt("ref_mode");
		if (ref_mode == -1)
		{
			int refCount = NGM.Stats->TotalSeqs = SequenceProvider.GetRefCount();
			int totalLen = 0;
			for (int i = 0; i < refCount; ++i)
				totalLen += SequenceProvider.GetRefLen(i);
			NGM.Stats->TotalBases = totalLen;

			// Caching:

			int sn = 0;
			int pageBases = 0;
			while (sn < refCount)
			{
				CSInit.PrefixIteration(
					SequenceProvider.GetRefSequence(sn),
					NGM.Stats->CurrentSeqBases = SequenceProvider.GetRefLen(sn),
					&CS::BuildPrefixTable,
					&(NGM.Stats->CurrentSeq = sn));
				Log.Message("Finished ref %d", sn);
				lastSeqTotal += NGM.Stats->CurrentSeqBases;
				pageBases += NGM.Stats->CurrentSeqBases;

				++sn;

				if (useCache && ((pageBases > cachePageLimit) || (sn >= refCount)))
				{
					cache->AddPage(RefTable, cRefTableLen);
					ClearRefTable();
					pageBases = 0;
				}
			}
		}
		else
		{
			// Stats
			if (NGM.DualStrand())
			{
				ref_mode *= 2;
				NGM.Stats->TotalSeqs = 2;
				NGM.Stats->TotalBases = SequenceProvider.GetRefLen(ref_mode) + SequenceProvider.GetRefLen(ref_mode + 1);
			}
			else
			{
				NGM.Stats->TotalSeqs = 1;
				NGM.Stats->TotalBases = SequenceProvider.GetRefLen(ref_mode);
			}

			CSInit.PrefixIteration(
				SequenceProvider.GetRefSequence(ref_mode),
				NGM.Stats->CurrentSeqBases = SequenceProvider.GetRefLen(ref_mode),
				&CS::BuildPrefixTable,
				&(NGM.Stats->CurrentSeq = ref_mode));
			lastSeqTotal += NGM.Stats->CurrentSeqBases;
			Log.Message("Finished forward.");

			if (useCache && (NGM.Stats->CurrentSeqBases > cachePageLimit))
			{
				cache->AddPage(RefTable, cRefTableLen);
				ClearRefTable();
			}

			if (NGM.DualStrand())
			{
				CSInit.PrefixIteration(
					SequenceProvider.GetRefSequence(ref_mode + 1),
					NGM.Stats->CurrentSeqBases = SequenceProvider.GetRefLen(ref_mode + 1),
					&CS::BuildPrefixTable,
					&(NGM.Stats->CurrentSeq = (ref_mode + 1)));
				lastSeqTotal += NGM.Stats->CurrentSeqBases;
				Log.Message("Finished reverse.");
			}

			if (useCache)
			{
				cache->AddPage(RefTable, cRefTableLen);
				ClearRefTable();
			}
		}

		NGM.Stats->CSStage = 3;
		PackRefTable();
		NGM.Stats->CSStage = 4;

		/*Log.Message("Writing cache");
		CSCache * cache = CSCache::CreateCache("test.cache");
		Log.Message("Page 1");
		cache->AddPage(RefTable, cRefTableLen);
		Log.Message("Page 2");
		cache->AddPage(RefTable, cRefTableLen);
		delete cache;*/
		if (cache != 0)
			delete cache;
		Log.Message("CS Init complete (elapsed: %.3fs)", init_tmr.ET());
	}
}

CS::~CS()
{
	if (rTable != 0)
		delete[] rTable;
}


void CS::CheckRTable()
{
	for (int i = 0; i < cSrchTableLen; ++i)
	{
		//if (rTable[i] != 0)
		//	Log.Warning("Unclean rTable (qry = %i, elem = %i): loc = %i", m_CurrentQrySeq, i, rTable[i]->Location.m_Location);
		
		if (rTable[i].Location.m_Location != cInvalidLocation)
		{
			Log.Warning("Unclean rTable (qry = %i, elem = %i): loc = %i", m_CurrentSeq, i, rTable[i].Location.m_Location);
			rTable[i] = sInvalidLocationScore;
		}
	}
}

// Refentry ist eine Gruppe (MaxRefsPerEntry) an Positionen, an der ein bestimmter Pr�fix vorkommt
//RefEntry ** RefTable;

PrefixRefRelay * FindOrInsert(ulong prefix, bool insert)
{
	int prefixHash = CSHash(prefix);
	PrefixRefRelay * cur = RefTable[prefixHash];

	while (cur != 0 && cur->Prefix != prefix)
		cur = cur->Next;

	if (insert && (cur == 0))
	{
		cur = new PrefixRefRelay();
		cur->Prefix = prefix;
		cur->Next = RefTable[prefixHash];
		RefTable[prefixHash] = cur;
	}

	return cur;
}

void CS::SaveToRefTable(ulong prefix, SequenceLocation loc)
{
	PrefixRefRelay * relay = FindOrInsert(prefix, true);

	if (relay->Entry == 0)
	{
		relay->Entry = new RefEntry();
		NGM.Stats->UniquePrefixCount++;
		NGM.Stats->RefEntryCount++;
	}
	else if (relay->Entry->refCount >= RefEntry::MaxRefsPerEntry)
	{
		RefEntry * tmp = relay->Entry;
		relay->Entry = new RefEntry();
		relay->Entry->nextEntry = tmp;
		relay->Entry->refTotal = tmp->refTotal;
		NGM.Stats->RefEntryCount++;
	}

	relay->Entry->ref[relay->Entry->refCount++] = loc;
	++relay->Entry->refTotal;
}
RefEntry * CS::GetRefEntry(ulong prefix) const
{
	PrefixRefRelay * relay = FindOrInsert(prefix, false);
	return (relay != 0) ? relay->Entry : 0;
}

void CS::PackRefTable()
{
/*	for (int i = 0; i < cRefTableLen; ++i)
	{
		PrefixRefRelay * relay = RefTable[i];
		while (relay != 0)
		{
			if (relay->Entry->refTotal > relay->Entry->refCount)
			{
				RefEntry * packed = new RefEntry(relay->Entry->refTotal);

				while ()
			}

			relay = relay->Next;
		}
	}*/
}

int PrefixRefRelay::sICount = 0;
int RefEntry::sICount = 0;

void CS::ClearRefTable()
{
	Log.Green("Active nodes: Relay: %i Count: %i", PrefixRefRelay::sICount, RefEntry::sICount);
	for (int i = 0; i < cRefTableLen; ++i)
	{
		PrefixRefRelay * cur = RefTable[i];
		while (cur != 0)
		{
			PrefixRefRelay * toDelete = cur;
			cur = cur->Next;

			delete toDelete->Entry;
			delete toDelete;
		}
		RefTable[i] = 0;
	}
	if ((PrefixRefRelay::sICount != 0) || (RefEntry::sICount != 0))
		Log.Warning("Lost nodes: Relay: %i Count: %i", PrefixRefRelay::sICount, RefEntry::sICount);
}

void ValidateCache(int page)
{
	static int x = 100;
	bool printed = false;

	Log.Message("Validating cache for page %i", page);
	CSCache const * cache = CSCache::LoadCache("test.cache");
	cache->LoadPage(page);
	for (int i = 0; i < CS::cRefTableLen; ++i)
	{
		if (RefTable[i] != 0)
		{
			PrefixRefRelay * cur = RefTable[i];
			RefEntry * cacheentry = cache->GetRefEntry(cur->Prefix);

			if (cacheentry->refTotal != cur->Entry->refTotal)
				Log.Error("Prefix %x: cached %i locs, mem %i locs", cur->Prefix, cacheentry->refTotal, cur->Entry->refTotal);

			if (!printed && i == x + page)
			{
				Log.Warning("Prefix %x: cached %i locs, mem %i locs", cur->Prefix, cacheentry->refTotal, cur->Entry->refTotal);
				for (int j = 0; j < cacheentry->refCount; ++j)
				{
					Log.Message("c: <%i, %i>", cacheentry->ref[j].m_Location, cacheentry->ref[j].m_RefId);
				}
				Log.Message("---");
				RefEntry * rc = cur->Entry;
				while (rc != 0)
				{
					for (int j = 0; j < rc->refCount; ++j)
					{
						Log.Message("m: <%i, %i>", cacheentry->ref[j].m_Location, cacheentry->ref[j].m_RefId);
					}
					rc = rc->nextEntry;
				}
				printed = true;
			}
			delete cacheentry;
		}
	}
	Log.Message("Done");
}
