#ifndef __COMMON_H__
#define __COMMON_H__

#include <vector>

#include "Types.h"

#include "Log.h"
#include "Config.h"
#include "PlatformSpecifics.h"
#include "SequenceProvider.h"
#include "Buffer.h"
#include "NGMThreads.h"
#include "MappedRead.h"

#include "NGMStats.h"

#include "NGMTask.h"

#include "Aligner.h"

#ifdef _WIN32
#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>

#ifndef NDEBUG
#define DEBUG_NEW new(_NORMAL_BLOCK, __FILE__, __LINE__)
#define new DEBUG_NEW
#endif
#endif

#undef module_name
#define module_name "NGM"

#include "MapFails.h"

class _NGM
{
public:
	static _NGM & Instance();

	int const GetNextReadBatch(int batchSize, int & readId);
	inline bool Running() const { return m_ActiveThreads > 0; }
	inline bool DualStrand() const { return m_DualStrand; }
	inline bool Paired() const { return m_Paired; }
	inline int GetOutputFormat() const { return m_OutputFormat; }

	IAlignment * Aligner() const { return AlignmentDispatcher::Instance(); }

	_NGM();
	~_NGM();

	void StartThread( NGMTask * task, int cpu = -1 );

	void AddUnmappedRead( int readid, int reason );
	int GetUnmappedReadCount() const;
	void AddMappedRead( int readid );
	int GetMappedReadCount() const;

	void AquireOutputLock();
	void ReleaseOutputLock();

	void InitQuit();

	bool StageActive( int stage )
	{
		return GetStageThreadCount(stage) > 0;
	}
	bool SingleStageActive( int stage )
	{
		return m_StageThreadCount[stage] > 0;
	}
	int GetStageThreadCount( int stage );

	NGMStats * Stats;

	// Switch to a list accessed by Stage?
	Buffer<LocationScore*> bCSSW;	// Buffer CS -> SW
	Buffer<MappedRead*> bSWO;	// Buffer SW -> Output

	static char const * AppName;
private:
	static NGMTHREADFUNC ThreadFunc(void*);

	static void Init();

	friend void NGMTask::FinishStage();
	void FinishStage( int tid );
	void FinishThread( int tid );

	friend int main(int argc, char* argv[]);
	static _NGM * pInstance;
	static NGMOnceControl once_control;
	static const int cMaxThreads = 1024;
	static const int cMaxStage = 8;

	volatile int m_NextRead;
	volatile int m_ReadCount;
	volatile int m_ActiveThreads;
	int m_NextThread;
	bool m_DualStrand;
	bool const m_Paired;
	int const m_OutputFormat;

	NGMMutex m_Mutex;
	NGMMutex m_OutputMutex;
	int m_StageThreadCount[cMaxStage];
	NGMThread m_Threads[cMaxThreads];
	NGMTask * m_Tasks[cMaxThreads];
	NGMMutex m_UMRMutex;
	std::vector<int> m_UnmappedReads;
	volatile int m_MappedReads;

	void UpdateReadCount(int n);

	friend class _SequenceProvider;
};

#undef module_name
#define module_name 0

#define NGM _NGM::Instance()

void Fatal();
void Help();

#endif
