#include <stdlib.h>
#include <stdio.h>

#ifdef _WIN32
#include <conio.h>
#endif

#ifdef _WIN32
#include <crtdbg.h>
#endif

#include "NGM.h"

#include "SW.h"
#include "CS.h"
#include "Output.h"


#include "Timing.h"

#include "Debug.h"

#undef module_name
#define module_name "MAIN"

volatile bool Terminating = false;

#include <vector>
std::vector<NGMTask*> sTasks;
typedef std::vector<NGMTask*>::iterator sTaskItr;

void StartCS()
{
	int cs_threadcount = Config.GetInt("cs_threads");
	if (cs_threadcount < 1)
	{
		Log.Warning("Illegal number of CS threads specified (%i) - defaulting to 1", cs_threadcount);
		cs_threadcount = 1;
	}
	int * cpu_affinities = new int[cs_threadcount];
	bool fixCPUs = Config.HasArray("cs_threads");

	if (fixCPUs)
		Config.GetIntArray("cs_threads", cpu_affinities, cs_threadcount);

	for (int i = 0; i < cs_threadcount; ++i)
	{
		CS * cs = new CS();
		NGM.StartThread(cs, (fixCPUs) ? cpu_affinities[i] : -1);
		sTasks.push_back(cs);
	}
	delete[] cpu_affinities;
}

void StartSWO()
{
	int sw_threadcount = Config.GetInt("sw_threads");
	if (sw_threadcount < 1)
	{
		Log.Warning("Illegal number of SW threads specified (%i) - defaulting to 1", sw_threadcount);
		sw_threadcount = 1;
	}

	for (int i = 0; i < sw_threadcount; ++i)
	{
		SW * sw = new SW();
		NGM.StartThread(sw);
		sTasks.push_back(sw);
	}

	Output * o = new Output(Config.GetString("output"));
	NGM.StartThread(o);
	sTasks.push_back(o);

}

void CheckOutput()
{
	if (!(Config.Exists("overwrite") && (Config.GetInt("overwrite") == 1)))
	{
		if (FileExists(Config.GetString("output")))
		{
			Log.Error("Output file already exists and overwrite set to false.");
			Fatal();
		}
	}
}

#ifdef _DEBUG
	bool cDebug = true;
#else
	bool cDebug = false;
#endif
IConfig * _config = 0;

#ifdef _WIN32
void _check()
{
	_CrtDumpMemoryLeaks();
}
#endif

int main(int argc, char * argv[])
{
#ifdef _WIN32
	_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	_CrtSetReportMode(_CRT_WARN, _CRTDBG_MODE_DEBUG);
#endif
	try
	{
		char const * arch[] = { "x86", "x64" };
		char const * build = (cDebug) ? " (DEBUG)" : "";
		Log.Message("Startup : %s%s (build %s %s)", arch[sizeof(void*) / 4 - 1], build, __DATE__, __TIME__);
		Timer tmr;
		tmr.ST();

		_NGM::AppName = argv[0];

		InitPlatform();

		// Initialization:
		//_Config::Init(argc, argv);	// Parses command line & parameter file
		_config = new _Config(argc, argv);
		_Log::Init();				// Inits logging to file

		CheckOutput();

		if (Config.Exists("master_cpu"))
			NGMSetThreadAffinity(0, Config.GetInt("master_cpu"));

		NGM;						// Init Core
		SequenceProvider.Init();	// Prepares input data

		((_Config*)_config)->Override("qry_max_len", SequenceProvider.GetReadMaxLen());
		Log.Message("Core initialization complete");

		CS::Init();					// Prepare CS (-> move into CS threads?)

		StartCS();
		StartSWO();

		// Main program loop
		int i = 0;
		float loads[2] = {0, 0};
		//_Log::FilterLevel(1);
#ifdef _WIN32
		char const * const loadbar = "\xFE\xFE\xFE\xFE\xFE\xFE\xFE\xFE\xFE\xFE";
#else
		char const * const loadbar = "##########";
#endif
		while (NGM.Running())
		{
			Sleep(50);
			while (!Terminating && _kbhit())
			{
				char ch = _getch();
				if ( ch == 'q')
					NGM.InitQuit();
			}
			loads[0] *= 0.25;
			loads[0] += (NGM.bCSSW.Load() * 0.75f);
			loads[1] /= 2;
			loads[1] += (NGM.bSWO.Load() / 2);

			if (++i == 10)
			{
	//			ClearConsole();
//				Log.Warning("Buffer status: bCSSW=%5.1f bSWO=%5.1f - %-10s", 100*loads[0], 100*loads[1], loadbar+10-(int)(loads[0]*10));
				i = 0;
			}
		}

		Log.Message("Done (%i reads mapped (%.2f%%), %i reads not mapped)(elapsed: %fs)", NGM.GetMappedReadCount(), (float)NGM.GetMappedReadCount() * 100.0f / (float)(NGM.GetMappedReadCount()+NGM.GetUnmappedReadCount()),NGM.GetUnmappedReadCount(), tmr.ET());
	}
	catch (...)
	{
		Log.Error("Unhandled exception in control thread");
		throw;
	}

	CS::Cleanup();
	_SequenceProvider::Cleanup();
	delete _config;
	_Log::Cleanup();

	delete _NGM::pInstance;

	CleanupPlatform();

#ifdef _WIN32
	atexit(_check);
#endif

	return 0;
}

void Fatal()
{
	Terminating = true;
	Log.Error("This error is fatal. Quitting...");
	ResetConsole();
	exit(1);
}

void Help()
{
	char const * const help_msg = "\
./NGM [parameter]\n\
\n\
Parameters:\n\
\n\
-p filename	Parameterfile (alternative: \"-c parameterfile\")\n\
		This Parameter is required.\n\
			\n\
-h		Prints help and aborts program\n\
-d		Creates a default config file and aborts program\n\
\n\
\n\
-i filename	Readfile	(alternative: \"-c:qry filename\")\n\
-r filename	Referenzfile	(alternative: \"-c:ref filename\")\n\
-o filename	Outputfile	(alternative: \"-c:output filename\")\n\
-l filename	Logfile		(alternative: \"-c:logfile filename -c:log 1\")\n\
\n\
-c:name value	Sets config entry \"name\" to \"value\" (example: -c:score_match 2)\n\n";

	printf(help_msg);
	exit(0);
}

// actually platform specific.../care
ulong const FileSize(char const * const filename)
{
	FILE * fp = fopen(filename, "rb");
	if (fp == 0)
	{
		Log.Warning("Tried to get size of nonexistant file %s", filename);
		return 0;
	}

	if (fseek (fp, 0, SEEK_END) != 0)
		return 0;

	ulong end = ftell(fp);
	fclose(fp);
	return end;
}

